#!/usr/bin/python3

import sys
import re
import tempfile
import subprocess
import argparse
import json
import pprint
import os
import shutil
import itertools
import urllib.request

MINIDUMP_TO_JSON = "/opt/backtrace/bin/minidump_to_json"
PROBE_DIRS = ("minidump", "symbols")

# location of symbol directories
# these are getting probed for symbol files along with
# the universe and project directories, if applicable
SYMBOLD_LOCATIONS = [
    "/var/symbold/symbols",
    "/backtrace/var/symbold/symbols",
]

def get_modules(dumpfile):
    cmd = [MINIDUMP_TO_JSON, dumpfile, "-"]
    if args.v:
        print(f"[-v]: capturing modules from {cmd}")
    m = subprocess.run(cmd, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if m.stderr:
        raise Exception(str(m.stderr.decode().strip()))
    m = json.loads(m.stdout)["modules"]
    m = {(e["debug_identifier"] or e["code_identifier"], e["debug_file"]) for e in m}
    return m

def resolve_relative_symbol_dirs(dump):
    paths = []
    dir_itr = os.path.dirname(os.path.abspath(dump))
    while True:
        for s_p in PROBE_DIRS:
            p = os.path.join(dir_itr, s_p)
            if os.path.isdir(p):
                paths.append(p)
        if os.path.samefile(dir_itr, os.path.join(dir_itr, "..")):
            break
        dir_itr = os.path.abspath(os.path.join(dir_itr, ".."))
    return paths

# this is a heuristic routine to try to find possible files
def resolve_symbol_files(dump, modules, symbol_dirs):
    symbol_files = []
    for mod in modules:
        for symbol_dir in symbol_dirs:
            symb_file = re.sub(r"\.pdb$", ".symb", mod[1])
            s_f = os.path.join(symbol_dir, mod[1], mod[0], symb_file)
            if os.path.exists(s_f):
                dest = os.path.join(outdir, "symbols", mod[1], mod[0], symb_file)
                symbol_files.append((s_f, dest))
                break
        else:
            missing_symbols.append(mod)
    return symbol_files, missing_symbols

def render_description(dump, symbol_dirs, out=sys.stdout):
    dirs = itertools.chain(*map(lambda d: ["--sym", d], symbol_dirs))
    cmd = [MINIDUMP_TO_JSON, *dirs, dump, "-"]
    if args.v:
        print(f"[-v]: capturing modules from {cmd}")
    m = subprocess.run(cmd, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if m.stderr:
        raise Exception(str(m.stderr.decode().strip()))
    report = json.loads(m.stdout)

    main_thread = report["mainThread"]
    crash_stack = report["threads"][main_thread]["stack"]
    modules = {}
    for m in report["modules"]:
        modules[m["debug_identifier"]] = m
        modules[m["code_file"]] = m

    print("{:<16}{:<16}{:<16}{:<16}{:<16}{:<16}{:<16}{}\t\t{}".format("frame_trust", "abs_address", "rel_address",
                                                                  "esp", "esp_gap", "ebp", "ebp_gap", "function...", "(module)"), file=out)
    print("="*(16*10), file=out);
    last = {}
    for frame in crash_stack:
        address = int(frame["address"], base=0)
        key = frame.get("debug_identifier") or frame.get("library")
        rel_address = address - modules.get("key", {}).get("start", 0)
        esp = int(frame.get("registers", {}).get("esp") or "0", base=0)
        ebp = int(frame.get("registers", {}).get("ebp") or "0", base=0)
        esp_gap = esp - (last.get("esp") or 0)
        ebp_gap = ebp - (last.get("ebp") or 0)
        func_name = frame.get("funcName", '')
        module = "library: `{}` debug_id: `{}`".format(frame.get("library", ""), frame.get("debug_identifier", ""))
        print(f"{frame.get('_frame_trust'):<16}{address:<16x}{rel_address:<16x}{esp:<16x}{esp_gap:<16x}{ebp:<16x}{ebp_gap:<16x}{func_name}\t\t[{module}]",
              file=out)
        last["esp"] = esp
        last["ebp"] = ebp

def read_symbold_config(symbold_config_file):
    with open(symbold_config_file) as symbold_config:
        return json.load(symbold_config)["symbol_server"]["servers"]

def download_symbols(ms):
    symbol_server = read_symbold_config(args.symbold_config)
    for server in symbol_server:
        if "aws_s3" in server:
            continue
        for s in ms:
            url = "{server}/{pdb}/{debug_id}/{pdb}".format(
                server=server["url"], pdb=s[1], debug_id=s[0])
            dest = os.path.join(outdir, "symbols", s[1], s[0], s[1])
            print(f"[-v]: downloading {url}")
            try:
                with urllib.request.urlopen(url) as request:
                    os.makedirs(os.path.dirname(dest))
                    with open(dest, "wb") as d:
                        shutil.copyfileobj(request, d)
            except urllib.error.HTTPError:
                print("[-v]:\t...skipping")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Crash analysis tool.')
    parser.add_argument('--list-local-symbol-files', action='store_true')
    parser.add_argument('-v', action='store_true')
    parser.add_argument('-a', '--archive',
                        help="Creates an archive which stores the dump as well as all symbol files.")
    parser.add_argument('--description', action='store_true',
                        help="Creates a short description of the minidump with symbols.")
    parser.add_argument('--use-symbolserver', action='store_true',
                        help="Contact symbol servers for missing symbols")
    parser.add_argument('--symbold-config')
    parser.add_argument('--outdir', '-o')
    parser.add_argument('dumpfile', nargs='+')
    args = parser.parse_args()

    if args.outdir:
        os.makedirs(args.outdir)
        outdir = args.outdir
    else:
        outdir = tempfile.mkdtemp()

    symbol_dirs = SYMBOLD_LOCATIONS[:]
    symbol_files = []
    missing_symbols = []
    for dump in args.dumpfile:
        crash_modules = get_modules(dump)
        symbol_dirs += resolve_relative_symbol_dirs(dump)
        sf, ms = resolve_symbol_files(dump, crash_modules, symbol_dirs)
        symbol_files += sf
        missing_symbols += ms

    if args.list_local_symbol_files:
        for s in symbol_files:
            print(s[0])

    if args.use_symbolserver:
        download_symbols(missing_symbols)

    if args.description:
        for dump in args.dumpfile:
            render_description(dump, symbol_dirs)

    if args.archive:
        for dump in args.dumpfile:
            shutil.copy(dump, args.outdir)
            with open(os.path.join(args.outdir, os.path.basename(dump) + ".txt"), "w") as report_file:
                render_description(dump, symbol_dirs, out=report_file)
        for s in symbol_files:
            os.makedirs(os.path.dirname(s[1]))
            shutil.copy(s[0], s[1])
        shutil.make_archive(args.archive, 'tar', outdir)

    if not args.outdir:
        shutil.rmtree(outdir)
